#pragma once

template <typename TType>
CMedian<TType>::CMedian(double fCMPPrecision)
{
	m_fCMPPrecision = fCMPPrecision;
}

template <typename TType>
CMedian<TType>::~CMedian()
{
}

template <typename TType>
void CMedian<TType>::SetDataset(const TTypeVctr &v)
{
	if(!v.size())
		throw std::runtime_error("GetMedianClassic invalid dataset!");

	m_DataSet = v;
}

template <typename TType>
void CMedian<TType>::AddDataset(const TTypeVctr &v)
{
	if(!v.size())
		throw std::runtime_error("GetMedianClassic invalid dataset!");

	m_DataSet.insert(m_DataSet.end(), v.begin(), v.end());
}

template <typename TType>
void CMedian<TType>::AddDatasetPoint(const TType &v)
{
	m_DataSet.push_back(v);
}

template <typename TType>
TType CMedian<TType>::CalcMedian(bool bDumpProfileData)
{
	if(!m_DataSet.size())
		throw std::runtime_error("Dataset not set!");

	m_nSelectRecursionCount = 0;
	m_nSelectRecursionCountMax = 0;
	m_nSelectCallsCount = 0;
	m_nTopSelectIterations = 0;

	size_t nMedianIndex = SelectTop(0, m_DataSet.size() - 1, m_DataSet.size() / 2);
	if(nMedianIndex >= m_DataSet.size())
		throw std::runtime_error("Internal error!");

	if(bDumpProfileData)
		printf("Select algorithm [of %lu data]: Total select calls : %lu, Top iterations : %lu, Recursion level : %lu\n", m_DataSet.size(), m_nSelectCallsCount + m_nTopSelectIterations, m_nTopSelectIterations, m_nSelectRecursionCountMax);

	assert(nMedianIndex == m_DataSet.size() / 2);	// should be what we are looking for
	TType Median = m_DataSet[nMedianIndex];

	// if element count even, get mean of other middle element and calc average
	if((m_DataSet.size() & 1) == 0)
	{
		// since we know all values less than our find "Median" are in lower part of our array now (because of previous partitioning)
		// And we are looking for biggest number of that set, go look for it
		size_t nRight = m_DataSet.size() / 2 - 1;
		typename TTypeVctr::iterator it_max = std::max_element(m_DataSet.begin(), m_DataSet.begin() + nRight + 1);
		Median = (Median + *it_max) / 2;
	}


	return Median;	
}

// semi-sort values between left & equal & right
template <typename TType>
size_t CMedian<TType>::Partition(size_t left, size_t right, size_t pivotIndex, size_t n)
{
	if(left >= m_DataSet.size() || right >= m_DataSet.size() || pivotIndex >= m_DataSet.size() || n >= m_DataSet.size())
		throw std::runtime_error("Partition() out of bounds!");

	TType pivotValue = m_DataSet[pivotIndex];
	std::swap(m_DataSet[pivotIndex], m_DataSet[right]);  // Move pivot to end
	size_t storeIndex = left;

	// Move all elements smaller than the pivot to the left of the pivot
	for(size_t i = left; i < right; i++)
	{
		// if list[i] < pivotValue
		if(pivotValue - m_DataSet[i] >= m_fCMPPrecision)			// if more than one precision distance
			std::swap(m_DataSet[storeIndex++], m_DataSet[i]);
	}

	// Move all elements equal to the pivot right after
	//the smaller elements
	size_t storeIndexEq = storeIndex;
	for(size_t i = storeIndex; i < right; i++)
	{
		if(std::abs(m_DataSet[i] - pivotValue) < (m_fCMPPrecision / 2)) // if within precision bounds
			std::swap(m_DataSet[storeIndexEq++], m_DataSet[i]);
	}

	// Move pivot to its final place
	std::swap(m_DataSet[right], m_DataSet[storeIndexEq]);

	// Return location of pivot considering the desired location n
	if(n < storeIndex)
		return storeIndex;  // n is in the group of smaller elements

	if(n <= storeIndexEq)
		return n;  // n is in the group equal to pivot

	return storeIndexEq; // n is in the group of larger elements
}

template <typename TType>
size_t CMedian<TType>::Select(size_t left, size_t right, size_t n)
{
	if(left >= m_DataSet.size() || right >= m_DataSet.size() || n >= m_DataSet.size())
		throw std::runtime_error("Select() out of bounds!");

m_nSelectCallsCount++;
m_nSelectRecursionCount++;
if(m_nSelectRecursionCount > m_nSelectRecursionCountMax)
	m_nSelectRecursionCountMax = m_nSelectRecursionCount;

//	size_t szData = right - left + 1;
//	printf("[%lu]\n", szData);

	size_t pivotIndex = -1;
	while(true)
	{
		if(left == right)
		{
m_nSelectRecursionCount--;
			return left;
		}

		pivotIndex = FindPivotIndex(left, right);		   // actual median of medians algorithm
		pivotIndex = Partition(left, right, pivotIndex, n);

		if(n == pivotIndex)
		{
m_nSelectRecursionCount--;
			return n;
		}
		else if(n < pivotIndex)
			right = pivotIndex - 1;
		else
			left = pivotIndex + 1;
	}
}

// return pivot number index
template <typename TType>
size_t CMedian<TType>::FindPivotIndex(size_t left, size_t right)
{
	if(left >= m_DataSet.size() || right >= m_DataSet.size())
		throw std::runtime_error("FindPivotIndex() out of bounds!");

	// for 5 or less elements just get median
	if(right - left < 5)
		return Partition5(left, right);

	// otherwise move the medians of five-element subgroups to the first n/5 positions
	size_t nDestIndex = left;
	for(size_t i = left; i <= right; i += 5)
	{
		// get the median position of the i'th five-element subgroup
		size_t subRight = i + 4;
		if(subRight > right)
			subRight = right;

		size_t median5 = Partition5(i, subRight);

		// swap list[median5] and list[left + floor((i - left)/5)]
		std::swap(m_DataSet[nDestIndex++], m_DataSet[median5]);
	}

	// compute the median of the n/5 medians-of-five
	size_t mid = (right - left) / 10 + left + 1;
	return Select(left, left + ((right - left) / 5), mid);
}

// TODO: Perhaps better to use decision tree & is not necessary to sort items 
// Actual sort of 5 elements & return median index
template <typename TType>
size_t CMedian<TType>::Partition5(size_t left, size_t right)
{
	if(left >= m_DataSet.size() || right >= m_DataSet.size())
		throw std::runtime_error("Partition5() out of bounds!");

//	std::sort(&m_DataSet[left], &m_DataSet[right] + 1);

	size_t i = left + 1;
	while(i <= right)
	{
		size_t j = i;
		while((j > left) && (m_DataSet[j-1] > m_DataSet[j]))
		{
			std::swap(m_DataSet[j-1], m_DataSet[j]);
			j--;
		}

		i++;
	}

	// floor((left + right) / 2);
	return (left + right) / 2;
}

template <typename TType>
size_t CMedian<TType>::SelectTop(size_t left, size_t right, size_t n)
{
	size_t pivotIndex = -1;
	while(true)
	{
		if(left == right)
			return left;

		pivotIndex = FindPivotIndex(left, right);
		pivotIndex = Partition(left, right, pivotIndex, n);
		m_nTopSelectIterations++;

		if(n == pivotIndex)
			return n;
		else if(n < pivotIndex)
			right = pivotIndex - 1;
		else
			left = pivotIndex + 1;
	}
}
